#include <avr/io.h>
#include <util/delay.h>

#include "usb_keyboard.h"

#define L_CTRL_INDEX	0
#define L_SHIFT_INDEX	1
#define L_ALT_INDEX		2
#define L_GUI_INDEX		3
#define R_CTRL_INDEX	4
#define R_SHIFT_INDEX	5
#define R_ALT_INDEX		6
#define R_GUI_INDEX		7

#define K_A					0x0004
#define K_B					0x0005
#define K_C					0x0006
#define K_D					0x0007
#define K_E					0x0008
#define K_F					0x0009
#define K_G					0x000a
#define K_H					0x000b
#define K_I					0x000c
#define K_J					0x000d
#define K_K					0x000e
#define K_L					0x000f
#define K_M					0x0010
#define K_N					0x0011
#define K_O					0x0012
#define K_P					0x0013
#define K_Q					0x0014
#define K_R					0x0015
#define K_S					0x0016
#define K_T					0x0017
#define K_U					0x0018
#define K_V					0x0019
#define K_W					0x001a
#define K_X					0x001b
#define K_Y					0x001c
#define K_Z					0x001d
#define K_1					0x001e
#define K_2					0x001f
#define K_3					0x0020
#define K_4					0x0021
#define K_5					0x0022
#define K_6					0x0023
#define K_7					0x0024
#define K_8					0x0025
#define K_9					0x0026
#define K_0					0x0027
#define K_ENTER			0x0028
#define K_ESC				0x0029
#define K_BACKSPACE	0x002a
#define K_TAB				0x002b
#define K_SPACE			0x002c
#define K_MINUS			0x002d
#define K_EQUALS		0x002e
#define K_L_SQUARE	0x002f
#define K_R_SQUARE	0x0030
#define K_BACKSLASH	0x0031
#define K_INT_HASH	0x0032
#define K_SEMICOLON	0x0033
#define K_QUOTE			0x0034
#define K_BACKQUOTE	0x0035
#define K_COMMA			0x0036
#define K_DOT				0x0037
#define K_SLASH			0x0038
#define K_CAPS_LOCK	0x0039
#define K_F1				0x003a
#define K_F2				0x003b
#define K_F3				0x003c
#define K_F4				0x003d
#define K_F5				0x003e
#define K_F6				0x003f
#define K_F7				0x0040
#define K_F8				0x0041
#define K_F9				0x0042
#define K_F10				0x0043
#define K_F11				0x0044
#define K_F12				0x0045
#define K_PRT_SCR		0x0046
#define K_SCRL_LOCK	0x0047
#define K_PAUSE			0x0048
#define K_INSERT		0x0049
#define K_HOME			0x004a
#define K_PAGE_UP		0x004b
#define K_DELETE		0x004c
#define K_END				0x004d
#define K_PAGE_DOWN	0x004e
#define K_RIGHT			0x004f
#define K_LEFT			0x0050
#define K_DOWN			0x0051
#define K_UP				0x0052
#define K_NUM_LOCK	0x0053
#define K_KP_SLASH	0x0054
#define K_KP_STAR		0x0055
#define K_KP_MINUS	0x0056
#define K_KP_PLUS		0x0057
#define K_KP_ENTER	0x0058
#define K_KP_1			0x0059
#define K_KP_2			0x005a
#define K_KP_3			0x005b
#define K_KP_4			0x005c
#define K_KP_5			0x005d
#define K_KP_6			0x005e
#define K_KP_7			0x005f
#define K_KP_8			0x0060
#define K_KP_9			0x0061
#define K_KP_0			0x0062
#define K_KP_DOT		0x0063
#define K_INT_BSLSH	0x0064
#define K_APP				0x0065
#define K_POWER			0x0066
#define K_KP_EQUALS	0x0067
#define K_F13				0x0068
#define K_F14				0x0069
#define K_F15				0x006a
#define K_F16				0x006b
#define K_F17				0x006c
#define K_F18				0x006d
#define K_F19				0x006e
#define K_F20				0x006f
#define K_F21				0x0070
#define K_F22				0x0071
#define K_F23				0x0072
#define K_F24				0x0073
#define K_EXECUTE		0x0074
#define K_HELP			0x0075
#define K_MENU			0x0076
#define K_SELECT		0x0077
#define K_STOP			0x0078
#define K_AGAIN			0x0079
#define K_UNDO			0x007a
#define K_CUT				0x007b
#define K_COPY			0x007c
#define K_PASTE			0x007d
#define K_FIND			0x007e
#define K_MUTE			0x007f
#define K_VOL_UP		0x0080
#define K_VOL_DOWN	0x0081
#define K_LOCK_CAPS	0x0082
#define K_LOCK_NUM	0x0083
#define K_LOCK_SCRL	0x0084
#define K_KP_COMMA	0x0085
#define K_KP_EQ2		0x0086
#define K_INT_1			0x0087
#define K_INT_2			0x0088
#define K_INT_3			0x0089
#define K_INT_4			0x008a
#define K_INT_5			0x008b
#define K_INT_6			0x008c
#define K_INT_7			0x008d
#define K_INT_8			0x008e
#define K_INT_9			0x008f
#define K_LANG_1		0x0090
#define K_LANG_2		0x0091
#define K_LANG_3		0x0092
#define K_LANG_4		0x0093
#define K_LANG_5		0x0094
#define K_LANG_6		0x0095
#define K_LANG_7		0x0096
#define K_LANG_8		0x0097
#define K_LANG_9		0x0098
#define K_ALT_ERASE	0x0099
#define K_SYS_RQ		0x009a
#define K_CANCEL		0x009b
#define K_CLEAR			0x009c
#define K_PRIOR			0x009d
#define K_RETURN		0x009e
#define K_SEPARATOR	0x009f
#define K_OUT				0x00a0
#define K_OPERATION	0x00a1
#define K_CLEAR_AGN	0x00a2
#define K_CR_SEL		0x00a3
#define K_EX_SEL		0x00a4

#define M_L_CTRL		(0x0100 | L_CTRL_INDEX)
#define M_L_SHIFT		(0x0100 | L_SHIFT_INDEX)
#define M_L_ALT			(0x0100 | L_ALT_INDEX)
#define M_L_GUI			(0x0100 | L_GUI_INDEX)
#define M_R_CTRL		(0x0100 | R_CTRL_INDEX)
#define M_R_SHIFT		(0x0100 | R_SHIFT_INDEX)
#define M_R_ALT			(0x0100 | R_ALT_INDEX)
#define M_R_GUI			(0x0100 | R_GUI_INDEX)

#define S_EXCL			(0x0200 | K_1)
#define S_AT				(0x0200 | K_2)
#define S_HASH			(0x0200 | K_3)
#define S_DOLLAR		(0x0200 | K_4)
#define S_PERCENT		(0x0200 | K_5)
#define S_HAT				(0x0200 | K_6)
#define S_AMP				(0x0200 | K_7)
#define S_STAR			(0x0200 | K_8)
#define S_L_BRACKET	(0x0200 | K_9)
#define S_R_BRACKET	(0x0200 | K_0)
#define S_USCORE		(0x0200 | K_MINUS)
#define S_PLUS			(0x0200 | K_EQUALS)
#define S_L_CURLY		(0x0200 | K_L_SQUARE)
#define S_R_CURLY		(0x0200 | K_R_SQUARE)
#define S_VBAR			(0x0200 | K_BACKSLASH)
#define S_INT_TILDE	(0x0200 | K_INT_HASH)
#define S_COLON			(0x0200 | K_SEMICOLON)
#define S_DQUOTE		(0x0200 | K_QUOTE)
#define S_TILDE			(0x0200 | K_BACKQUOTE)
#define S_LESS			(0x0200 | K_COMMA)
#define S_GREATER		(0x0200 | K_DOT)
#define S_QUESTION	(0x0200 | K_SLASH)
#define S_INT_VBAR	(0x0200 | K_INT_BSLSH)

#define L_HOLD(X)		(0x0300 | (X))

#define L_LOCK(X)		(0x0400 | (X))


#define NUM_LAYERS 2
#define NUM_ROWS 5
#define NUM_COLUMNS 16
#define MAX_LAYER_STACK_SIZE 4

#define NUM_DEBOUNCE_STATES 2


static const uint16_t KEY_MAP[NUM_LAYERS][NUM_ROWS][NUM_COLUMNS] = {
	{	{0,						K_1,					K_2,			K_3,				K_4,			K_5,				0,						0,					0,						0,						K_6,					K_7,			K_8,				K_9,					K_0,					0},
		{K_ESC,				K_QUOTE,			K_COMMA,	K_DOT,			K_P,			K_Y,				K_L_SQUARE,		0,					0,						K_R_SQUARE,		K_F,					K_G,			K_C,				K_R,					K_L,					K_EQUALS},
		{K_BACKQUOTE,	K_A,					K_O,			K_E,				K_U,			K_I,				K_TAB,				0,					0,						K_ENTER,			K_D,					K_H,			K_T,				K_N,					K_S,					K_MINUS},
		{K_BACKSLASH,	K_SEMICOLON,	K_Q,			K_J,				K_K,			K_X,				0,						0,					0,						0,						K_B,					K_M,			K_W,				K_V,					K_Z,					K_SLASH},
		{0,						0,						0,				M_L_SHIFT,	M_L_ALT,	M_L_CTRL,		M_L_SHIFT,		L_HOLD(1),	K_BACKSPACE,	K_SPACE,			M_L_CTRL,			M_L_ALT,	M_L_SHIFT,	0,						0,						0}},
	{	{0,						K_F1,					K_F2,			K_F3,				K_F4,			K_F5,				0,						0,					0,						0,						K_F6,					K_F7,			K_F8,				K_F9,					K_F10,				0},
		{K_ESC,				0,						0,				0,					0,				0,					S_L_BRACKET,	0,					0,						S_R_BRACKET,	K_PAGE_UP,		K_HOME,		K_UP,				K_END,				K_APP,				K_F11},
		{K_PRT_SCR,		0,						0,				M_L_SHIFT,	M_L_CTRL,	K_0,				K_TAB,				0,					0,						K_ENTER,			K_PAGE_DOWN,	K_LEFT,		K_DOWN,			K_RIGHT,			M_L_GUI,			K_F12},
		{K_CAPS_LOCK,	S_EXCL,				S_AT,			S_HASH,			S_DOLLAR,	S_PERCENT,	0,						0,					0,						0,						S_HAT,				S_AMP,		S_STAR,			S_L_BRACKET,	S_R_BRACKET,	K_SCRL_LOCK},
		{0,						0,						0,				M_L_SHIFT,	M_L_ALT,	M_L_CTRL,		M_L_SHIFT,		L_HOLD(1),	K_DELETE,			K_INSERT,			M_L_CTRL,			M_L_ALT,	M_L_SHIFT,	0,						0,						0}}};


static uint16_t columnStates[NUM_ROWS] = {0};
static uint16_t debounceColumnStates[NUM_ROWS][NUM_DEBOUNCE_STATES] = {{0}};
static uint8_t debounceStateIndex = 0;

static uint8_t keyDownLayers[NUM_ROWS][NUM_COLUMNS] = {{0}};

static uint8_t modifierCounts[8] = {0};
static uint8_t numKeysDown = 0;

static uint8_t layerHoldStack[MAX_LAYER_STACK_SIZE] = {0};
static uint8_t layerHoldStackSize = 0;
static uint8_t layerLockStack[MAX_LAYER_STACK_SIZE] = {0};
static uint8_t layerLockStackSize = 0;
static uint8_t layer = 0;



static void DeActivateRows(void)
{
	DDRF &= 0b10001100;
	PORTF &= 0b10001100;
}

static void ActivateRow(uint8_t rowIndex)
{
	uint8_t rowMask;

	DeActivateRows();
	switch (rowIndex)
	{	case 0:
		case 1:
			rowMask = 1 << rowIndex;
			DDRF |= rowMask;
			PORTF &= ~rowMask;
			break;
		case 2:
		case 3:
		case 4:
			rowMask = 1 << (rowIndex + 2);
			DDRF |= rowMask;
			PORTF &= ~rowMask;
			break;
	}
}

static uint16_t ReadColumn(void)
{
	uint16_t columnState = 0;
	columnState |= ((uint16_t)~PINB) & 0b0000000011111111;
	columnState |= ((uint16_t)~PINC << 2) & 0b0000001100000000;
	columnState |= ((uint16_t)~PIND << 10) & 0b1111110000000000;
	return columnState;
}

int main(void)
{
	uint8_t i, j, k;

	CLKPR = 0x80;
	CLKPR = 0;

	DDRB &= 0b00000000;
	PORTB |= 0b11111111;
	DDRC &= 0b00111111;
	PORTC |= 0b11000000;
	DDRD &= 0b11000000;
	PORTD |= 0b00111111;

	DeActivateRows();

	usb_init();
	while (1)
	{	if (usb_configured())
			break;
	}
	_delay_ms(1000);

	while (1)
	{	uint8_t keysChanged = 0;
		for (i=0; i<NUM_ROWS; i++)
		{	ActivateRow(i);
			_delay_us(30);
			uint16_t currentColumnState = ReadColumn();
			uint16_t *rowDebounceColumnStates = debounceColumnStates[i];
			for (j=0; j<NUM_COLUMNS; j++)
			{	uint16_t columnMask = 1 << j;
				uint16_t keyState = currentColumnState & columnMask;
				uint8_t keyStateDifferent = 0;
				for (k=0; k<NUM_DEBOUNCE_STATES; k++)
				{	if (keyState != (rowDebounceColumnStates[k] & columnMask))
					{	keyStateDifferent = 1;
						break;
				}	}
				if (!keyStateDifferent)
				{	if (keyState)
					{	// Key down
						if (!(columnStates[i] & columnMask))
						{	// Key pressed
							columnStates[i] = columnStates[i] | columnMask;
							keyDownLayers[i][j] = layer;
							uint16_t keyId = KEY_MAP[layer][i][j];
							if (keyId != 0)
							{	uint8_t keyType = (uint8_t)(keyId >> 8);
								uint8_t keyCode = (uint8_t)keyId;
								uint8_t foundLayerLockKey;
								switch (keyType)
								{	case 0:
										// Character
										if (numKeysDown < 6)
										{	keyboard_keys[numKeysDown] = keyCode;
											numKeysDown++;
											keysChanged = 1;
										}
										break;
									case 1:
										// Modifier
										if (modifierCounts[keyCode] <= 0)
										{	uint8_t modifierCode = 1 << keyCode;
											keyboard_modifier_keys |= modifierCode;
											keysChanged = 1;
										}
										modifierCounts[keyCode]++;
										break;
									case 2:
										// Shifted character
										if (numKeysDown < 6)
										{	if (modifierCounts[L_SHIFT_INDEX] <= 0)
											{	uint8_t modifierCode = 1 << L_SHIFT_INDEX;
												keyboard_modifier_keys |= modifierCode;
											}
											modifierCounts[L_SHIFT_INDEX]++;
											keyboard_keys[numKeysDown] = keyCode;
											numKeysDown++;
											keysChanged = 1;
										}
										break;
									case 3:
										// Layer hold
										if (layerHoldStackSize < MAX_LAYER_STACK_SIZE)
										{	layerHoldStack[layerHoldStackSize] = keyCode;
											layerHoldStackSize++;
											layer = keyCode;
										}
										break;
									case 4:
										// Layer lock
										foundLayerLockKey = 0;
										for (k=0; k<layerLockStackSize; k++)
										{	if (layerLockStack[k] == keyCode)
											{	uint8_t kNext = k + 1;
												while (kNext < layerLockStackSize)
												{	layerLockStack[k] = layerLockStack[kNext];
													k = kNext;
													kNext = kNext + 1;
												}
												layerLockStackSize--;
												if (layerHoldStackSize <= 0)
												{	if (layerLockStackSize > 0)
														layer = layerLockStack[layerLockStackSize - 1];
													else
														layer = 0;
												}
												foundLayerLockKey = 1;
												break;
										}	}
										if (!foundLayerLockKey)
										{	if (layerLockStackSize < MAX_LAYER_STACK_SIZE)
											{	layerLockStack[layerLockStackSize] = keyCode;
												layerLockStackSize++;
												if (layerHoldStackSize <= 0)
													layer = keyCode;
										}	}
										break;
					}	}	}	}
					else
					{	// Key up
						if (columnStates[i] & columnMask)
						{	// Key released
							columnStates[i] = columnStates[i] & ~columnMask;
							uint16_t keyId = KEY_MAP[keyDownLayers[i][j]][i][j];
							if (keyId != 0)
							{	uint8_t keyType = (uint8_t)(keyId >> 8);
								uint8_t keyCode = (uint8_t)keyId;
								switch (keyType)
								{	case 0:
										// Character
										for (k=0; k<numKeysDown; k++)
										{	if (keyboard_keys[k] == keyCode)
											{	uint8_t kNext = k + 1;
												while (kNext < numKeysDown)
												{	keyboard_keys[k] = keyboard_keys[kNext];
													k = kNext;
													kNext = kNext + 1;
												}
												keyboard_keys[k] = 0;
												numKeysDown--;
												keysChanged = 1;
												break;
										}	}
										break;
									case 1:
										// Modifier
										modifierCounts[keyCode]--;
										if (modifierCounts[keyCode] <= 0)
										{	uint8_t modifierCode = 1 << keyCode;
											keyboard_modifier_keys &= ~modifierCode;
											keysChanged = 1;
										}
										break;
									case 2:
										// Shifted character
										for (k=0; k<numKeysDown; k++)
										{	if (keyboard_keys[k] == keyCode)
											{	uint8_t kNext = k + 1;
												while (kNext < numKeysDown)
												{	keyboard_keys[k] = keyboard_keys[kNext];
													k = kNext;
													kNext = kNext + 1;
												}
												keyboard_keys[k] = 0;
												numKeysDown--;
												modifierCounts[L_SHIFT_INDEX]--;
												if (modifierCounts[L_SHIFT_INDEX] <= 0)
												{	uint8_t modifierCode = 1 << L_SHIFT_INDEX;
													keyboard_modifier_keys &= ~modifierCode;
												}
												keysChanged = 1;
												break;
										}	}
										break;
									case 3:
										// Layer hold
										for (k=0; k<layerHoldStackSize; k++)
										{	if (layerHoldStack[k] == keyCode)
											{	uint8_t kNext = k + 1;
												while (kNext < layerHoldStackSize)
												{	layerHoldStack[k] = layerHoldStack[kNext];
													k = kNext;
													kNext = kNext + 1;
												}
												layerHoldStackSize--;
												if (layerHoldStackSize > 0)
													layer = layerHoldStack[layerHoldStackSize - 1];
												else if (layerLockStackSize > 0)
													layer = layerLockStack[layerLockStackSize - 1];
												else
													layer = 0;
												break;
										}	}
										break;
				}	}	}	}	}
				rowDebounceColumnStates[debounceStateIndex] =
						(rowDebounceColumnStates[debounceStateIndex] & ~columnMask) | keyState;
		}	}
		debounceStateIndex = (debounceStateIndex + 1) % NUM_DEBOUNCE_STATES;

		if (keysChanged)
			usb_keyboard_send();

//$
//maybe make this shorter, because I'm doing the debouncing for 2 iterations
//and there's also the time it takes to run the loop here
//maybe do some debug output to show how much actual time has elapsed between iterations
		_delay_ms(5);
	}

	return 0;
}
